"""
The `pytoil remove` subcommand group.

Author: Tom Fleet
Created: 18/06/2021
"""

import shutil
from typing import List

import typer
from wasabi import msg

from pytoil.cli import utils
from pytoil.config import Config

app = typer.Typer(name="remove", no_args_is_help=True)


# Callback for documentation only
@app.callback()
def remove() -> None:
    """
    Remove projects from your local filesystem.

    The remove command provides an easy interface for decluttering your local
    projects directory.

    You can selectively remove any number of projects or nuke the whole thing
    if you want.

    As with most programmatic deleting, the directories are deleted instantly and
    not sent to trash. As such, pytoil will prompt you for confirmation before
    doing anything.

    The "--force/-f" flag can be used to force deletion without the confirmation
    prompt. Use with caution!
    """


# Again, can't call it 'all', python keyword
@app.command(name="all")
def all_(
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force delete without confirmation.",
        show_default=False,
    )
) -> None:
    """
    Delete all your projects.

    Recursively delete every subdirectory in your configured
    projects directory.

    This is really reserved for those "I'm having a really bad day" moments
    and should be treated with caution.

    You will be prompted for confirmation unless the '--force/-f' flag is used.

    Examples:

    $ pytoil remove all

    $ pytoil remove all --force
    """
    config = Config.from_file()

    local_projects = utils.get_local_projects(path=config.projects_dir)

    if not local_projects:
        msg.warn("You don't have any local projects to remove!", spaced=True, exits=1)

    if not force:
        typer.confirm(
            "This will remove ALL of your projects. Are you sure?", abort=True
        )

    # If we get here, user either used '--force' or said yes

    for project in sorted(local_projects, key=str.casefold):
        typer.secho(f"Removing project: {project!r}.", fg=typer.colors.YELLOW)
        shutil.rmtree(config.projects_dir.joinpath(project))

    msg.good("Done!", spaced=True)


@app.command()
def these(
    projects: List[str] = typer.Argument(..., help="Name of the project(s) to delete."),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force delete without confirmation.",
        show_default=False,
    ),
) -> None:
    """
    Delete specified projects.

    Recursively delete the specified projects from your configured
    projects directory.

    You will be prompted for confirmation unless the '--force/-f' flag is used.

    Examples:

    $ pytoil remove these project1 project2 project

    $ pytoil remove these project1 project2 project3 --force
    """
    config = Config.from_file()

    local_projects = utils.get_local_projects(path=config.projects_dir)

    if not local_projects:
        msg.warn("You don't have any local projects to remove!", spaced=True, exits=1)

    # If user gives a project that isn't here (e.g. typo), abort
    for project in projects:
        if project not in local_projects:
            msg.warn(
                f"{project!r} not found in projects directory. Was it a typo?",
                spaced=True,
                exits=1,
            )

    if not force:
        if len(projects) <= 5:
            # A manageable length to display each one
            typer.confirm(
                f"This will remove {', '.join(projects)} from your local filesystem. Are you sure?",  # noqa: E501
                abort=True,
            )
        else:
            # Too many to show and look nice, just show a number
            typer.confirm(
                f"This will remove {len(projects)} projects from your local filesystem. Are you sure?",  # noqa: E501
                abort=True,
            )

    # If we get here, user either used '--force' or said yes to the prompt

    for project in sorted(projects, key=str.casefold):
        typer.secho(f"Removing project: {project!r}.", fg=typer.colors.YELLOW)
        shutil.rmtree(config.projects_dir.joinpath(project))

    msg.good("Done!", spaced=True)



"""
Module responsible for creating and managing
python virtual environments through the std lib `venv` module.

Author: Tom Fleet
Created: 20/06/2021
"""

import subprocess
import venv
from pathlib import Path
from typing import List, Optional

from pytoil.environments import Environment
from pytoil.exceptions import MissingInterpreterError


class Venv(Environment):
    def __init__(self, project_path: Path) -> None:
        """
        Representation of a virtualenv.

        A VirtualEnv's `.executable` property (also a Path) points
        to the python executable of the newly created virtualenv. This
        executable may or may not exist, and the existence check is
        the primary means of checking whether or not this virtualenv
        exists.

        Note: It is important not to resolve `.executable` as it could
        resolve back to the system python if it is a symlink.

        Args:
            project_root (Path): The root path of the current project.
        """
        self._project_path = project_path.resolve()

    def __repr__(self) -> str:
        return self.__class__.__qualname__ + f"(project_path={self.project_path!r})"

    @property
    def project_path(self) -> Path:
        return self._project_path

    @property
    def executable(self) -> Path:
        return self._project_path.joinpath(".venv/bin/python")

    def exists(self) -> bool:
        """
        Checks whether the virtual environment exists by a proxy
        check if the `executable` exists.

        If this executable exists then both the project and virtual environment
        must also exist and must be valid.
        """
        return self.executable.exists()

    def create(self, packages: Optional[List[str]] = None) -> None:
        """
        Create the virtual environment in the project.

        If packages are specified here, these will be installed
        once the environment is created.

        Args:
            packages (Optional[List[str]], optional): Packages to install immediately
                after environment creation. Defaults to None.
        """
        # `clear` will ensure any existing venv is destroyed first rather
        # than causing an error
        # TODO: `create()` has an `upgrade_deps` arg to do what `update_seeds` does but
        # it's only in python 3.9
        # Add this in later
        venv.create(
            env_dir=self.project_path.joinpath(".venv"), clear=True, with_pip=True
        )

        # Update core deps
        self.update_seeds()

        # Install any specified packages
        if packages:
            self.install(packages=packages)

    def update_seeds(self) -> None:
        """
        Venv will install 'seed' packages to a new
        environment: `pip`, `setuptools` and `wheel`.

        It is good practice to keep these packages fully up to date
        this method does exactly that by invoking pip from
        the virtualenvs executable.

        This is equivalent to running:
        `python -m pip install --upgrade pip setuptools wheel`
        from the command line with the virtualenv activated.
        """

        # Validate the executable
        if not self.exists():
            raise MissingInterpreterError(f"Interpreter: {self.executable} not found.")

        try:
            # We don't need to specify a cwd here because `self.executable` is
            # an absolute path
            subprocess.run(
                [
                    f"{self.executable}",
                    "-m",
                    "pip",
                    "install",
                    "--upgrade",
                    "--quiet",
                    "pip",
                    "setuptools",
                    "wheel",
                ],
                check=True,
            )

        except subprocess.CalledProcessError:
            raise

    def install(self, packages: List[str]) -> None:
        """
        Generic `pip install` method.

        Takes a list of packages to install. All packages are passed through
        to pip so any versioning syntax will work as expected.

        Args:
            packages (List[str]): A list of valid packages to install.
                Analogous to simply calling `pip install *packages` from within
                the virtualenv. If only 1 package, still must be in a list e.g.
                `["black"]`.
        """

        cmd: List[str] = [f"{self.executable}", "-m", "pip", "install", "--quiet"]

        cmd.extend(packages)

        try:
            subprocess.run(cmd, check=True)
        except subprocess.CalledProcessError:
            raise



"""
The `pytoil config` subcommand group.

Author: Tom Fleet
Created: 18/06/2021
"""

import time

import typer
from wasabi import msg

from pytoil.config import Config, defaults

app = typer.Typer(name="config", no_args_is_help=True)

# If the file doesn't exist, let's make a default one nicely
# This will check if the config file exists
# every time the `pytoil` command is run so should
# eliminate those nasty FileNotFoundErrors
if not defaults.CONFIG_FILE.exists():
    msg.warn("No pytoil config file detected!")

    with msg.loading("Making you a starter file..."):
        Config.helper().write()
        time.sleep(1)

    msg.good(f"Your file is now available at {defaults.CONFIG_FILE}", spaced=True)
    msg.text("Try running pytoil again!", spaced=True, exits=0)


# Callback for documentation only
@app.callback()
def config() -> None:
    """
    Interact with pytoil's configuration.

    The config command group allows you to get, set, and show pytoil's configuration.
    Getting and showing obviously do not edit the config file ($HOME/.pytoil.yml).

    Setting a key value pair using the 'config set' subcommand will cause the config
    file to be updated and overwritten. You will be prompted to confirm this however.

    Currently the only config key you cannot set on the command line is
    'common_packages'. If you want to change this, you'll have to do so in the
    config file itself.
    """


@app.command()
def show() -> None:
    """
    Show pytoil's config.

    The show command allows you to easily see pytoil's current config.

    The values are taken directly from the config file where specified or
    the defaults otherwise.
    """
    config = Config.from_file()

    typer.secho("\nPytoil Config:\n", fg=typer.colors.CYAN, bold=True)

    for key, val in config.to_dict().items():
        typer.echo(f"{key}: {val!r}")


@app.command()
def get(
    key: str = typer.Argument(..., help="Config key to fetch the value for.")
) -> None:
    """
    Get the currently set value for a config key.

    The get command only allows valid pytoil config keys.
    """
    if key not in defaults.CONFIG_KEYS:
        msg.warn(f"{key!r} is not a valid pytoil config key.", exits=1)

    config = Config.from_file().to_dict()
    typer.secho(f"\n{key}: {config.get(key)!r}\n")


@app.command()
def set(
    key: str = typer.Argument(..., help="Config key to set the value for."),
    val: str = typer.Argument(..., help="Value to set for <key>."),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force config overwrite without confirmation.",
        show_default=False,
    ),
) -> None:
    """
    Set a config key, value pair.

    The set command allows you to set a config value.

    On confirmation, the config file ($HOME/.pytoil.yml) will be
    overwritten with the new data.

    The '--force/-f' flag can be used to force overwrite without confirmation.
    """
    if key not in defaults.CONFIG_KEYS:
        msg.warn(f"{key!r} is not a valid pytoil config key.", exits=1)

    # Can't set a list via the command line
    if key == "common_packages":
        msg.warn("You can't set common_packages at the command line", exits=1)

    config = Config.from_file().to_dict()

    new_setting = {key: val}

    config.update(new_setting)  # type: ignore

    new_config = Config.from_dict(config)

    if not force:
        typer.confirm(f"This will set {key!r} to {val}. Are you sure?", abort=True)

    try:
        new_config.write()
    except Exception:
        # This should only happen if something really weird happens
        # like corrupted config file etc.
        msg.fail(
            "Uh oh! Something went wrong! Check your config file for errors.",
            spaced=True,
            exits=1,
        )
    else:
        msg.good("Config successfully updated", spaced=True)
        msg.text(f"{key!r} is now {config[key]}")  # type: ignore


@app.command(name="help")
def help_() -> None:
    """
    Print a list and description of pytoil config keys.
    """
    msg.divider("The '.pytoil.yml' config file")
    print(defaults.CONFIG_SCHEMA)



"""
Module implementing pytoils interface with VSCode.

Author: Tom Fleet
Created: 19/06/2021
"""

import json
import shutil
import subprocess
from pathlib import Path
from typing import Any, Dict, Optional

from pytoil.exceptions import CodeNotInstalledError


class VSCode:
    def __init__(self, root: Path, code: Optional[str] = shutil.which("code")) -> None:
        self.root = root
        self.code = code
        self.workspace_settings = self.root.joinpath(".vscode/settings.json")

    def __repr__(self) -> str:
        return self.__class__.__qualname__ + f"(root={self.root!r}, code={self.code!r})"

    def raise_for_code(self) -> None:
        if not self.code:
            raise CodeNotInstalledError(
                "VSCode binary application `code` not found on $PATH."
            )

    def open(self) -> None:
        """
        Opens a project root in VSCode using the
        `code` CLI command.
        """

        self.raise_for_code()

        try:
            # Use the code binary to open the project
            subprocess.run([f"{self.code}", f"{self.root}"], check=True)
        except subprocess.CalledProcessError:
            raise

    def set_python_path(self, python_path: Path) -> None:
        """
        Sets the VSCode workspace setting `python.pythonPath`
        to `python_path`, overwriting if already exists.

        Take care not to resolve `python_path` as it will
        resolve the symlink back to the system python. As such,
        `python_path` should be the executable from a `VirtualEnv`
        or `CondaEnv` class to ensure this is safely done.

        Args:
            python_path (pathlib.Path): The path of the
                virtual environments python interpreter.
        """

        new_settings_dict: Dict[str, str] = {"python.pythonPath": str(python_path)}

        if not self.workspace_settings.exists():
            self.workspace_settings.parent.mkdir(parents=True)
            self.workspace_settings.touch()

        if len(self.workspace_settings.read_text()) == 0:
            # The file is empty, we don't have to worry about
            # preserving it's contents
            with open(self.workspace_settings, mode="w", encoding="utf-8") as f:
                json.dump(new_settings_dict, f, sort_keys=True, indent=4)
        else:
            # File exists and is not empty, let's preserve whatever
            # settings are already here
            with open(self.workspace_settings, mode="r", encoding="utf-8") as f:
                settings: Dict[str, Any] = json.load(f)

            # Wipe the file and recreate it so it's now empty
            self.workspace_settings.unlink()
            self.workspace_settings.touch()

            # Update the settings with our new python path
            settings.update(new_settings_dict)

            # Write the new settings back
            with open(self.workspace_settings, mode="w", encoding="utf-8") as f:
                json.dump(settings, f, sort_keys=True, indent=4)



"""
Module responsible for interacting with git via
subprocesses.

Author: Tom Fleet
Created: 19/06/2021
"""

import shutil
import subprocess
from pathlib import Path
from typing import Optional

from pytoil.config import defaults
from pytoil.exceptions import GitNotInstalledError


class Git:
    def __init__(self, git: Optional[str] = shutil.which("git")) -> None:
        self.git = git

    def __repr__(self) -> str:
        return self.__class__.__qualname__ + f"(git={self.git!r})"

    def raise_for_git(self) -> None:
        if not self.git:
            raise GitNotInstalledError(
                "'git' executable not found on $PATH. Is git installed?"
            )

    def run(
        self, *args: str, check: bool = True, cwd: Path = defaults.PROJECTS_DIR
    ) -> None:
        """
        Generic method to run `git` in a subprocess.

        Pass any args to git via *args.

        e.g. 'run("clone", "https://github.com/me/repo.git")'

        Args:
            check (bool, optional): Will raise a CalledProcessError if
                something goes wrong. Defaults to True.

            cwd (Path, optional): Working directory of the subprocess.
                Defaults to defaults.PROJECTS_DIR.

        Raises:
            subprocess.CalledProcessError: If the subprocess command fails.
        """

        self.raise_for_git()

        try:
            subprocess.run([f"{self.git}", *args], check=check, cwd=cwd)
        except subprocess.CalledProcessError:
            raise

    def clone(
        self, url: str, check: bool = True, cwd: Path = defaults.PROJECTS_DIR
    ) -> None:
        """
        Convenience wrapper around `self.run` to clone a repo.

        Args:
            url (str): The clone url of the repo.

            check (bool, optional): Raise CalledProcessError on failure.
                Defaults to True.

            cwd (Path, optional): Working directory of the subprocess.
                Defaults to defaults.PROJECTS_DIR.

        Raises:
            subprocess.CalledProcessError: If the subprocess command fails.
        """

        self.run("clone", url, check=check, cwd=cwd)

    def init(self, path: Path, check: bool = True) -> None:
        """
        Convenience wrapper around `self.run` to initialise a repo.

        Args:
            path (Path): Root of the project to initialise
                the repo in.

            check (bool, optional): Raise CalledProcessError on failure.
                Defaults to True.

        Raises:
            subprocess.CalledProcessError: If the subprocess command fails.
        """

        self.run("init", check=check, cwd=path)



"""
Assorted helper functions for the pytoil CLI.

Author: Tom Fleet
Created: 19/06/2021
"""

from pathlib import Path
from typing import List, Set

import httpx
from cookiecutter.main import cookiecutter
from wasabi import msg

from pytoil.api import API
from pytoil.config import Config
from pytoil.environments import Conda, Environment, Venv
from pytoil.exceptions import EnvironmentAlreadyExistsError
from pytoil.git.git import Git
from pytoil.repo import Repo


def get_local_projects(path: Path) -> Set[str]:
    """
    Returns all the projects (directories) under
    `path`.
    """

    return {f.name for f in path.iterdir() if f.is_dir() and not f.name.startswith(".")}


def warn_if_no_api_creds(config: Config) -> None:
    """
    Will print a helpful warning message and exit the program
    if username or token are not filled out in the config file.
    """

    if not config.can_use_api():
        msg.warn(
            "You must fill set your username and token to use API features!",
            spaced=True,
            exits=1,
        )


def pre_new_checks(repo: Repo, api: API) -> None:
    """
    Checks whether the repo already exists either locally
    or remotely, prints helpful warning messages and exits
    the program if True.
    """

    is_local = repo.exists_local()

    try:
        is_remote = repo.exists_remote(api=api)
    except httpx.HTTPStatusError as err:
        handle_http_status_errors(error=err)
    else:

        if is_local:
            msg.warn(
                title=f"{repo.name!r} already exists locally!",
                text=f"To checkout this project, use 'pytoil checkout {repo.name}'.",
                spaced=True,
                exits=1,
            )
        elif is_remote:
            msg.warn(
                title=f"{repo.name!r} already exists on GitHub!",
                text=f"To checkout this project, use 'pytoil checkout {repo.name}'.",
                spaced=True,
                exits=1,
            )


def make_new_project(
    repo: Repo, git: Git, cookie: str, use_git: bool, config: Config
) -> None:
    """
    Create a new development project either from a cookiecutter
    template or from scratch.
    """
    if cookie:
        # We don't initialise a git repo for cookiecutters
        # some templates have hooks which do this, mine do!
        msg.info(f"Creating {repo.name!r} from cookiecutter: {cookie!r}.")
        cookiecutter(template=cookie, output_dir=config.projects_dir)

    else:
        msg.info(f"Creating {repo.name!r} at {repo.local_path}.")
        # Make an empty dir and git repo
        repo.local_path.mkdir(parents=True)
        if use_git:
            git.init(path=repo.local_path, check=True)


def create_virtualenv(repo: Repo, packages: List[str]) -> Environment:
    """
    Creates and returns new virtual environment with packages and reports
    to user.
    """

    msg.info(
        f"Creating virtual environment for {repo.name!r}",
        text=f"Including packages: {', '.join(packages)}",
        spaced=True,
    )
    env = Venv(project_path=repo.local_path)
    with msg.loading("Working..."):
        env.create(packages=packages)

    return env


def create_condaenv(repo: Repo, packages: List[str]) -> Environment:
    """
    Creates and returns new conda environment with packages and reports
    to user.
    """
    msg.info(
        f"Creating conda environment for {repo.name!r}",
        text=f"Including packages: {', '.join(packages)}",
        spaced=True,
    )
    env = Conda(name=repo.name, project_path=repo.local_path)
    try:
        with msg.loading("Working..."):
            env.create(packages=packages)
    except EnvironmentAlreadyExistsError:
        msg.warn(
            f"Conda environment {env.name!r} already exists!", spaced=True, exits=1
        )

    return env


def handle_http_status_errors(error: httpx.HTTPStatusError) -> None:
    """
    Handles a variety of possible HTTP Status errors, print's nicer output
    to the user, and exits the program if necessary.

    Call this in an except block on CLI commands accessing the
    GitHub API.

    Args:
        error (httpx.HTTPStatusError): The error to be handled.
    """
    code = error.response.status_code

    if code == 401:
        msg.fail(
            title="HTTP Error: 401 - Unauthorized",
            text="This usually means something is wrong with your token!",
            exits=1,
        )
    elif code == 404:
        msg.fail(
            title="HTTP Error: 404 - Not Found",
            text="This usually means something is up with the GitHub API.",
            exits=1,
        )
    elif code == 500:
        msg.fail(
            title="HTTP Error: 500 - Internal Server Error",
            text="This usually means GitHub is not happy.",
            exits=1,
        )



"""
The pytoil app CLI command and it's direct child commands.

Author: Tom Fleet
Created: 18/06/2021
"""


from enum import Enum
from typing import List

import httpx
import typer
from wasabi import msg

from pytoil import __version__
from pytoil.api import API
from pytoil.cli import config, pull, remove, show, utils
from pytoil.config import Config
from pytoil.exceptions import EnvironmentAlreadyExistsError, RepoNotFoundError
from pytoil.git import Git
from pytoil.repo import Repo
from pytoil.vscode import VSCode

PYTOIL_DOCS_URL: str = "https://followtheprocess.github.io/pytoil/"

# Create the root app
app = typer.Typer(name="pytoil", no_args_is_help=True)

# Add the sub commands
app.add_typer(show.app, name="show")
app.add_typer(remove.app, name="remove")
app.add_typer(config.app, name="config")
app.add_typer(pull.app, name="pull")


# Choice of virtual environments for a new project
class VirtualEnv(str, Enum):
    """
    Choice of virtualenvs to create in a new project.
    """

    venv = "venv"
    conda = "conda"
    none = "none"


def version_callback(value: bool) -> None:
    if value:
        typer.echo(f"pytoil version: {__version__}")
        raise typer.Exit(0)


# Callback for documentation only
@app.callback()
def main(
    version: bool = typer.Option(
        None,
        "--version",
        "-V",
        help="Display pytoil version.",
        callback=version_callback,
        is_eager=True,
        show_default=False,
    )
) -> None:
    """
    Helpful CLI to automate the development workflow.

    - Create and manage your local and remote projects

    - Build projects from cookiecutter templates.

    - Easily create/manage virtual environments.

    - Minimal configuration required.
    """


@app.command()
def checkout(
    project: str = typer.Argument(
        ...,
        help="Name of the project to checkout.",
    ),
    venv: bool = typer.Option(
        False,
        "--venv",
        "-v",
        help="Attempt to auto-create a virtual environment.",
        show_default=False,
    ),
) -> None:
    """
    Checkout an existing development project.

    The checkout command lets you easily resume work on an existing project,
    whether that project is available locally in your configured projects
    directory, or if it is on GitHub.

    If pytoil finds your project locally, and you have enabled VSCode in your
    config file it will open it for you. If not, it will just tell you it already
    exists locally and where to find it.

    If your project is on GitHub, pytoil will clone it for you and then open it
    (or tell you where it cloned it if you dont have VSCode set up).

    Finally, if checkout can't find a match after searching locally and on GitHub,
    it will prompt you to use 'pytoil new' to create a new one.

    You can also ask pytoil to automatically create a virtual environment on
    checkout with the '--venv/-v' flag. This only happens for projects pulled down
    from GitHub to avoid accidentally screwing up local projects.

    If the '--venv/-v' flag is used, pytoil will look at your project to try and detect
    which type of environment to create (conda or standard python).

    More info can be found in the documentation.

    Examples:

    $ pytoil checkout my_project

    $ pytoil checkout my_project --venv
    """
    config = Config.from_file()
    utils.warn_if_no_api_creds(config)

    # Setup the objects required
    api = API(username=config.username, token=config.token)
    repo = Repo(
        owner=config.username,
        name=project,
        local_path=config.projects_dir.joinpath(project),
    )
    code = VSCode(root=repo.local_path)
    git = Git()

    is_local = repo.exists_local()
    try:
        is_remote = repo.exists_remote(api=api)
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_errors(error=err)
    else:

        if is_local:
            # No environment or git stuff here, chances are if it exists locally
            # user has already done all this stuff
            msg.info(f"{repo.name!r} available locally.", spaced=True)

            if config.vscode:
                msg.text(f"Opening {repo.name!r} in VSCode.")
                code.open()

        elif is_remote:
            msg.info(f"{repo.name!r} found on GitHub. Cloning...", spaced=True)
            git.clone(url=repo.clone_url, check=True, cwd=config.projects_dir)

            env = repo.dispatch_env()

            if venv:
                if not env:
                    msg.warn(
                        "Unable to auto-detect required environent. Skipping.",
                        spaced=True,
                    )
                else:
                    msg.info("Auto creating correct virtual environment.", spaced=True)

                    try:
                        with msg.loading("Creating Environment..."):
                            env.create(packages=config.common_packages)
                    except EnvironmentAlreadyExistsError:
                        msg.warn(
                            title="Environment already exists!",
                            text="No need to create a new one. Skipping.",
                        )
                    finally:
                        if config.vscode:
                            code.set_python_path(python_path=env.executable)

            if config.vscode:
                msg.info(f"Opening {repo.name!r} in VSCode.", spaced=True)
                code.open()

        else:
            msg.warn(
                title=f"{repo.name!r} not found locally or on GitHub!",
                text=f"Does it exist? If not, create a new project with 'pytoil new {repo.name}'.",  # noqa: E501
                exits=1,
            )


@app.command(context_settings={"allow_extra_args": True})
def new(
    ctx: typer.Context,
    project: str = typer.Argument(..., help="Name of the project to create."),
    cookie: str = typer.Option(
        None,
        "--cookie",
        "-c",
        help="URL to a cookiecutter template repo from which to build the project.",
    ),
    venv: VirtualEnv = typer.Option(
        VirtualEnv.none,
        "--venv",
        "-v",
        help="Which type of virtual environment to create for the project.",
        case_sensitive=False,
        show_default=True,
    ),
    no_git: bool = typer.Option(
        False,
        "--no-git",
        "-n",
        help="Don't initialise an empty git repo in the app of the project.",
    ),
) -> None:
    """
    Create a new development project.

    Bare usage will simply create an empty folder in your configured projects
    directory.

    You can also create a project from a cookiecutter template by passing a valid
    url to the '--cookie/-c' flag.

    By default, pytoil will initialise an empty git repo in the folder, following
    the style of modern language build tools such as rust's cargo. You can disable
    this behaviour by setting 'init_on_new' to false in pytoil's config file
    or by passing the '--no-git/-n' flag here.

    If you want pytoil to create a new virtual environment for your project, you
    can use the '--venv/-v' flag. Standard python and conda virtual environments
    are supported.

    If the '--venv/-v' flag is used, you may also pass a list of python packages
    to install into the created virtual environment. These will be delegated to
    the appropriate tool (pip or conda) depending on what environment was created.
    If the environment is conda, the packages will be passed at environment creation
    time meaning they will have their dependencies resolved together. Normal python
    environments will first be created and then have specified packages installed.

    If 'common_packages' is specified in pytoil's config file, these will automatically
    be included in the environment.

    To specify versions of packages via the command line, you must enclose them
    in double quotes e.g. "flask>=1.0.0" not flask>=1.0.0 otherwise this will
    be interpreted by the shell as a command redirection.

    Examples:

    $ pytoil new my_project

    $ pytoil new my_project --cookie https://github.com/some/cookie.git

    $ pytoil new my_project --venv conda

    $ pytoil new my_project -c https://github.com/some/cookie.git -v conda --no-git

    $ pytoil new my_project -v venv requests "flask>=1.0.0"
    """
    # Get config and ensure user can access API
    config = Config.from_file()
    utils.warn_if_no_api_creds(config)

    # Setup the objects required
    api = API(username=config.username, token=config.token)
    repo = Repo(
        owner=config.username,
        name=project,
        local_path=config.projects_dir.joinpath(project),
    )
    code = VSCode(root=repo.local_path)
    git = Git()

    if ctx.args:
        packages: List[str] = config.common_packages + ctx.args
    else:
        packages = config.common_packages

    # Resolve config vs flag for no-git
    # flag takes priority over config
    use_git: bool = config.init_on_new and not no_git

    # Check is project already exists and warn/exit if so
    utils.pre_new_checks(repo=repo, api=api)

    # If we get here, all is well and we can create stuff!
    utils.make_new_project(
        repo=repo, git=git, cookie=cookie, use_git=use_git, config=config
    )

    if venv.value == venv.venv:
        env = utils.create_virtualenv(repo=repo, packages=packages)

        if config.vscode:
            msg.info(f"Opening {repo.name!r} in VSCode.", spaced=True)
            code.set_python_path(env.executable)
            code.open()

    elif venv.value == venv.conda:
        env = utils.create_condaenv(repo=repo, packages=packages)

        if config.vscode:
            msg.info(f"Opening {repo.name!r} in VSCode.", spaced=True)
            code.set_python_path(env.executable)
            code.open()

    else:
        # Only other allowed condition is none
        typer.secho(
            "Virtual environment not requested. Skipping environment creation.",
            fg=typer.colors.YELLOW,
        )

        if config.vscode:
            msg.info(f"Opening {repo.name!r} in VSCode.", spaced=True)
            code.open()


@app.command()
def info(
    project: str = typer.Argument(..., help="Name of the project to fetch info for.")
) -> None:
    """
    Get useful info for a project.

    Given a project name (can be local or remote), 'info' will return a summary
    description of the project.

    If the project is on GitHub, info will prefer getting information from the GitHub
    API as this is more detailed.

    If the project is local only, some information is extracted from the operating
    system about the project.

    Examples:

    $ pytoil info my_project
    """
    config = Config.from_file()
    utils.warn_if_no_api_creds(config)

    api = API(username=config.username, token=config.token)

    repo = Repo(
        owner=config.username,
        name=project,
        local_path=config.projects_dir.joinpath(project),
    )

    try:
        _ = repo.exists_remote(api=api)
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_errors(error=err)

    try:
        with msg.loading(f"Fetching info for {project!r}..."):
            info_dict = repo.info(api=api)

    except RepoNotFoundError:
        msg.warn(
            f"{project!r} not found locally or on GitHub. Was it a typo?",
            spaced=True,
            exits=1,
        )
    else:
        typer.secho(f"\nInfo for {project!r}:\n", fg=typer.colors.CYAN, bold=True)
        for key, val in info_dict.items():
            typer.echo(f"{key}: {val}")


@app.command()
def docs() -> None:  # pragma: no cover
    """
    Open pytoil's documentation in your browser.

    Examples:

    $ pytoil docs
    """
    msg.info("Opening pytoil's docs in your browser...", spaced=True)
    typer.launch(url=PYTOIL_DOCS_URL)


@app.command()
def gh(
    project: str = typer.Argument(..., help="Name of the project to go to on GitHub.")
) -> None:  # pragma: no cover
    """
    Open one of your projects on GitHub.

    Given a project name (must exist on GitHub and be owned by you),
    'gh' will open your browser and navigate to the project on GitHub.

    Examples:

    $ pytoil gh my_project
    """
    config = Config.from_file()
    utils.warn_if_no_api_creds(config)

    api = API(username=config.username, token=config.token)

    repo = Repo(
        owner=config.username,
        name=project,
        local_path=config.projects_dir.joinpath(project),
    )

    try:
        exists_remote = repo.exists_remote(api=api)
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_errors(error=err)
    else:

        if not exists_remote:
            msg.warn(
                f"{project!r} not found on GitHub! Was it a typo?", spaced=True, exits=1
            )
        else:
            msg.info(f"Opening {project!r} on GitHub...", spaced=True)
            typer.launch(url=repo.html_url)



class PytoilException(Exception):
    """
    Base pytoil exception from which all subclasses
    must inherit.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class GitNotInstalledError(PytoilException):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class CodeNotInstalledError(PytoilException):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class MissingInterpreterError(PytoilException):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class CondaNotInstalledError(PytoilException):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class EnvironmentAlreadyExistsError(PytoilException):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class BadEnvironmentFileError(PytoilException):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class EnvironmentDoesNotExistError(PytoilException):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class UnsupportedCondaInstallationError(PytoilException):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class RepoNotFoundError(PytoilException):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)



"""
The `pytoil pull` subcommand group.

Author: Tom Fleet
Created: 18/06/2021
"""

from typing import List, Set

import httpx
import typer
from wasabi import msg

from pytoil.api import API
from pytoil.cli import utils
from pytoil.config import Config
from pytoil.git import Git
from pytoil.repo import Repo

app = typer.Typer(name="pull", no_args_is_help=True)


# Callback for documentation only
@app.callback()
def pull() -> None:
    """
    Pull down your remote projects.

    The pull command provides easy methods for pulling down remote projects.

    It is effectively a nice wrapper around git clone but you don't have to worry
    about urls or what your cwd is, pull will grab your remote projects by
    name and clone them to your configured projects directory.

    You can also use pull to batch clone multiple repos, even all of them if you're
    into that sorta thing.

    Any remote project that already exists locally will be skipped and
    none of your local projects are changed in any way.

    It's very possible to accidentally clone a lot of repos when using pull
    so you will be prompted for confirmation before pytoil does anything.

    The "--force/-f" flag can be used to override this confirmation prompt
    if desired.
    """


@app.command(name="all")
def all_(
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force pull without confirmation.",
    )
) -> None:
    """
    Pull down all your remote projects.

    This will look at the projects you already have locally and
    on GitHub, calculate the difference and then use git to clone
    each one to your configured projects directory.

    Any projects that already exist locally are completely skipped and
    will not be altered in any way.

    You will be prompted for confirmation unless the "--force/-f" flag
    is used.
    """
    config = Config.from_file()
    utils.warn_if_no_api_creds(config)

    api = API(username=config.username, token=config.token)

    try:
        with msg.loading("Calculating difference..."):
            local_projects = utils.get_local_projects(path=config.projects_dir)
            remote_projects = api.get_repo_names()

            diff = remote_projects.difference(local_projects)
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_errors(error=err)
    else:

        pull_diff(
            diff=diff, remotes=remote_projects, force=force, config=config, git=Git()
        )


@app.command()
def these(
    projects: List[str] = typer.Argument(
        ..., help="Name of the project(s) to pull down."
    ),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force pull without confirmation.",
    ),
) -> None:
    """
    Pull down specified remote projects.

    This will look at the projects you've specified, those you already have locally, and
    those on GitHub, calculate the difference and then use git to clone
    each one to your configured projects directory.

    Any projects that already exist locally are completely skipped and
    will not be altered in any way.

    You will be prompted for confirmation unless the "--force/-f" flag
    is used.
    """
    config = Config.from_file()
    utils.warn_if_no_api_creds(config)

    api = API(username=config.username, token=config.token)

    try:
        with msg.loading("Calculating difference..."):
            local_projects = utils.get_local_projects(path=config.projects_dir)
            remote_projects = api.get_repo_names()

            for project in projects:
                if project not in remote_projects:
                    msg.warn(
                        f"{project!r} not found on GitHub. Was it a typo?",
                        spaced=True,
                        exits=1,
                    )

            diff = set(projects).difference(local_projects)
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_errors(error=err)
    else:

        pull_diff(
            diff=diff, remotes=remote_projects, force=force, config=config, git=Git()
        )


def pull_diff(
    diff: Set[str], remotes: Set[str], force: bool, config: Config, git: Git
) -> None:
    """
    Helper function to help the pull CLI command with repeated logic.

    Args:
        diff (Set[str]): The difference to clone.
        remotes (Set[str]): Set of remote projects.
        force (bool): The value for the '--force' flag.
        config (Config): The Config object.
    """

    if not diff:
        msg.info("Your local and remote projects are in sync.")
        msg.good("Nothing to pull!", exits=0)

    if not remotes:
        msg.warn("You don't have any remote projects to pull!", spaced=True, exits=1)

    if not force:
        if len(diff) <= 5:
            # A managable length to display each one
            typer.confirm(
                f"This will pull down {', '.join(diff)}. Are you sure?", abort=True
            )
        else:
            # Too many to show to look nice, just show number
            typer.confirm(
                f"This will pull down {len(diff)} projects. Are you sure?", abort=True
            )

    # If we get here, we're good to go

    for project in diff:
        # Make a Repo object for each
        repo = Repo(
            owner=config.username,
            name=project,
            local_path=config.projects_dir.joinpath(project),
        )

        msg.info(f"Cloning {project!r}", spaced=True)
        git.clone(url=repo.clone_url, cwd=config.projects_dir)

    msg.good("Done!", spaced=True)



from pytoil.environments.abstract import Environment
from pytoil.environments.conda import Conda
from pytoil.environments.virtualenv import Venv

__all__ = [
    "Environment",
    "Venv",
    "Conda",
]
